
#include "aconfig.h"
#include "config.h"
#include "nntpd.h"
#include "setproctitle.h" 
#include "memory.h"
#include "aprotos.h"
#include "tgetopt.h"


struct option opts[] = {
    { "config", 1, NULL, 'c', 0 },
    { "help",   0, NULL, 'h', 0 },
    { "test",   0, NULL, 't', 0 },
    { NULL,     0, NULL, 0,   0 },
};

static const char* usage = "Usage: %s [options]\n"
    "   -c, --config CONFIGFILE\tSpecify configfile instead of searching for it.\n"
    "   -h, --help             \tCommand usage.\n"
    "   -t, --test             \tTest configuration and exit.\n";

int test_only = 0;

PROTO int writepidfile(void) {
    FILE *pidf;
    struct passwd *pwent;

    if ( (pidf = fopen(cfg.PidFile,"w")) == NULL ) {
	die("Can't write pidfile");
	return -1;
    }

    fprintf(pidf, "%d", (int)getpid());
    fclose(pidf);

    /* change ownership, so we can overwrite it later when we're forked
     * under a different user (probably news) */

    if ( (pwent=getpwnam(cfg.RunAsUser)) == NULL )
        die("unknown user %s", cfg.RunAsUser);
    chown(cfg.PidFile, pwent->pw_uid, pwent->pw_gid);

    return 0;
}


/* daemon() - detach process from user and disappear into the background  */
/* returns -1 on failure, but you can't do much except exit in that case, */
/* since we may already have forked. This is based on the BSD version,    */
/* so the caller is responsible for things like the umask, etc.           */

/* believed to work on all Posix systems */

pid_t init_daemon(void) {
    pid_t pid;
    switch ( (pid=fork()) )
    {
        case 0:  break;
        case -1: return -1;
        default: _exit(0);          /* exit the original process */
    }

    closeall(1);
    if (setsid() < 0)               /* shoudn't fail */
      return -1;

    /* dyke out this switch if you want to acquire a control tty in the */
    /* future - not normally advisable for daemons */

/*    switch ( fork() )
    {
        case 0:  break;
        case -1: return -1;
        default: close(2);
		 _exit(0);
    }
*/
    return pid;
}

/* This is the daemon's main work - listen for connections and spawn.  */

void run_daemon()
{
    struct sockaddr_in addr;
    struct hostent *hp = NULL;
    int addrlen = sizeof(addr);
    int flag = 1;
    pid_t pid;
    struct passwd *pwent;
    int sock, rc;
    CLIENT *initclient;
    int curconcur;
    char buf[128];
    ulong ipaddr;
    ACL *acl;

    loadconfig();

    close(0);
    close(1);
    
    if ( (sock = socket(PF_INET, SOCK_STREAM, 0) == -1) ) 
	die("Can't create socket");

    if ( setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(flag)) < 0 )
	die("Can't set socket options");

    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(cfg.ListenPort);
    addr.sin_addr.s_addr = inet_addr(cfg.BindAddress);

    if ( ! test_only) {
        if ( bind(sock, (struct sockaddr *) &addr, addrlen) < 0 ) 
	   die("Bind failed");
	if ( listen(sock, cfg.ListenBackLog) < 0 ) 
	    die("Listen failed");
	info("Listening on %s:%d", cfg.BindAddress, cfg.ListenPort);

	/* first time, as root, and chown to news */
	writepidfile();

	if ( getuid() == 0 ) {
	    if ( (pwent=getpwnam(cfg.RunAsUser)) == NULL )
		die("Unknown user %s, check configuration", cfg.RunAsUser);
	    if ( setuid(pwent->pw_uid) )
		die("Cant setuid %s", cfg.RunAsUser);
	}
	errno = 0;
    }

    load_servers();
    loadactive();
    loadoverviewfmt();

    loadprofiles(0);
    loadaccess();
    loadauth();

    if ( test_only ) {
	info("Startup Test Successfull, Exiting..");
	exit(0);
    }else
	info("NNTP Server Starting..");

    closelog();
    if ( (pid=init_daemon()) < 0)
        die("Can't fork");

    /* 2nd time, with the right pid, as user news */
    writepidfile();

    openlog("nntpswitchd", LOG_PID, LOG_NEWS);
    syslog(LOG_NOTICE, "Server running new pid %d uid %d euid %d", (int)getpid(), (int)getuid(), (int)geteuid() );

    for (;;)
    {
        rc = accept(sock, (struct sockaddr *) &addr, &addrlen);

        if (rc >= 0) {
	    if ( master->connections > cfg.MaxConnections ) { 
		write(rc, MSG_TOOMANY, strlen(MSG_TOOMANY));
		syslog(LOG_ERR,"Too many global users, increase cfg.MaxConnections");
		goto cleanup;
	    }
	    
	    ipaddr = ntohl(addr.sin_addr.s_addr);
	    if ( (acl = getacl(ipaddr)) == NULL ) {
		syslog(LOG_ERR,"cant get default ACL");
		write(rc, MSG_NOACL, strlen(MSG_NOACL));
		goto cleanup;
	    }

            if ( acl->deny ) {
		write(rc, MSG_NOPERM, strlen(MSG_NOPERM));
		syslog(LOG_NOTICE, "denied connection from %s", inet_ntoa(addr.sin_addr));
		goto cleanup;
	    }

 	    if ( (initclient = find_client(
				    ipaddr, 
				    &curconcur, 
				    acl->unlimit ? 0 : acl->profile->MaxConcurrent 
				)) == NULL ) {
		sprintf(buf,MSG_TOOMANY_CONCURRENT, acl->profile->MaxConcurrent);
		write(rc, buf, strlen(buf));
                syslog(LOG_ERR, "Too many concurrent connections from %s", inet_ntoa(addr.sin_addr));
		goto cleanup;
	    }

	    initclient->acl = acl;

	    switch (fork()) {
	    	case 0:
		    hp = gethostbyaddr((char*)&addr.sin_addr.s_addr, sizeof(addr.sin_addr.s_addr), AF_INET); 
		    close(sock);
		    initclient->socket = rc;
		    run_child(initclient, strdup(hp != NULL ? (char *)hp->h_name : inet_ntoa(addr.sin_addr)), ipaddr);
		    _exit(0);

		case -1: 
		    info("couldnt fork child %s", strerror(errno)); 
		    close(rc); 
		    break;
	    }

cleanup:
	    close(rc);
	}
    }
}


void chld_handler(int sig) {
    while ( waitpid( -1, 0, WNOHANG ) > 0 ) {}
}

void hup_handler(int sig) {
    syslog(LOG_ERR,"sighup received");
    reload_servers();
    reloadactive();
    reloadprofiles();
    reloadaccess();
    reloadauth();
    loadconfig();
}

void segv_handler(int sig) {
    syslog(LOG_ERR,"sigsegv received");
    if ( client != NULL ) 
        close_child_noacc(client);
    _exit(1);
}
void int_handler(int sig) {
    syslog(LOG_ERR,"sigint received");
    if ( client != NULL ) 
        close_child_noacc(client);
    _exit(1);
}
void term_handler(int sig) {
    syslog(LOG_ERR,"sigterm received");
    if ( client != NULL ) {
	client_quit(client);
        close_child(client);
    }
    _exit(1);
}
void ill_handler(int sig) {
    syslog(LOG_ERR,"sigill received");
    if ( client != NULL ) 
        close_child_noacc(client);
    _exit(1);
}
void abrt_handler(int sig) {
    syslog(LOG_ERR,"sigabrt received");
    if ( client != NULL ) 
        close_child_noacc(client);
    _exit(1);
}


int main(int argc, char **argv)
{
    struct sigaction sighup, sigsegv, sigpipe, sigint, sigterm, sigill, sigabrt;
    struct sigaction sigchld;

    tgetopt(argc, argv);

    if ( toptset('h') ) {
	printf(usage, argv[0]);
	exit(0);
    }

    if ( toptset('t') )
	test_only++;

    if ( toptset('c') )
	config_file = toptargs('c');

    memset(&sigchld,0,sizeof(sigchld));
    sigchld.sa_handler = &chld_handler;
    sigaction(SIGCHLD, &sigchld, NULL);
    
    memset(&sighup,0,sizeof(sighup));
    sighup.sa_handler = &hup_handler;
    sigaction(SIGHUP, &sighup, NULL);
    
    memset(&sigsegv,0,sizeof(sigsegv));
    sigsegv.sa_handler = &segv_handler;
    sigaction(SIGSEGV, &sigsegv, NULL);

    memset(&sigpipe,0,sizeof(sigpipe));
    sigpipe.sa_handler = SIG_IGN;
    sigaction(SIGPIPE, &sigpipe, NULL);

    memset(&sigint,0,sizeof(sigint));
    sigint.sa_handler = &int_handler;
    sigaction(SIGINT, &sigint, NULL);

    memset(&sigterm,0,sizeof(sigterm));
    sigterm.sa_handler = &term_handler;
    sigaction(SIGTERM, &sigterm, NULL);

    memset(&sigill,0,sizeof(sigill));
    sigill.sa_handler = &ill_handler;
    sigaction(SIGILL, &sigill, NULL);

    memset(&sigabrt,0,sizeof(sigabrt));
    sigabrt.sa_handler = &abrt_handler;
    sigaction(SIGABRT, &sigabrt, NULL);

    openlog("nntpswitchd", LOG_PID, LOG_NEWS);
    syslog(LOG_NOTICE,"Server starting up..");

    initproctitle(argc,argv);
    setproctitle("nntpswitchd: waiting for connections");

    if ( (master=memmap(sizeof(MASTER))) == NULL ) {
        die("Can't allocate master memory");
        exit(2);
    }

    run_daemon();
    return 0;
}

