/*
 * profile.c - Authentication and Accounting Profiles Configuration
 *
 * $Id: profile.c,v 1.7 2004/03/14 23:33:25 tommy Exp $
 */

#include "aconfig.h"
#include "nntpd.h"
#include "memory.h"
#include "aprotos.h"

/* Max size of config line entry */
#define BUFS 1024

int linenr;
PROFILE *profiles;
ALIAS *aliases;
WILDMAT *wildmats[MAX_WILDMATS];
const char *default_wildmat = "*";
                                                                                                                             

/*
 * General Configuration Functions
 */

PROTO void clear_wmat_aliases(void) {
    int i;

    /* Clear Wildmats */
    for (i=0; i<master->numwildmats; i++) {
        free(wildmats[i]->name);
	free(wildmats[i]->pattern);
	free(wildmats[i]);
    }
    info("Freed %d Wildmats", master->numwildmats);
    master->numwildmats = 0;

    /* Clear Aliases */
    memunmap(aliases, sizeof(ALIAS) * MAX_ALIASES);
    info("Freed %d Aliases", master->numaliases);
    master->numaliases = 0;

    return;
}

PROTO void clear_profiles(void) {
    /* Clear Profiles */
    memunmap(profiles, sizeof(PROFILE) * MAX_PROFILES);
    info("Freed %d Profiles", master->numprofiles);
    master->numprofiles = 0;

    return;
}


PROTO void reloadprofiles(void) {
    struct stat _stat;

    if ( stat(cfg.ProfileFile, &_stat) == -1 ) 
	die("Can't stat profile config file %s", cfg.ProfileFile);

    if ( master->laststatprofile < _stat.st_mtime ) {
	clear_wmat_aliases();
	loadprofiles(1);
    }
}


PROTO void loadprofiles(int reload) {
    FILE *f;
    struct stat _stat;
    char buf[BUFS];

    if ( reload == 0 ) {
        if ( (profiles=memmap(sizeof(PROFILE) * MAX_PROFILES)) == NULL )
	    die("Cant mmap profiles");
        memset(profiles, 0, sizeof(PROFILE) * MAX_PROFILES);
    }

    if ( (aliases=memmap(sizeof(ALIAS) * MAX_ALIASES)) == NULL )
	die("Cant mmap aliases");
    memset(aliases, 0, sizeof(ALIAS) * MAX_ALIASES);

    if ( (f=fopen(cfg.ProfileFile,"r")) == NULL )
	die("Can't open profile config %s", cfg.ProfileFile);

    linenr=0;
    while( fgets(buf, BUFS, f) ) {
	chop(buf);
	linenr++;

	if ( buf[0] != '\r' && buf[0] != '\n' && buf[0] != ';' && buf[0] != '#' && buf[0] != 0 ) {

	          if ( strncasecmp(buf, "profile", 7) == 0 ) {
		parsecfg_profile(buf, f, reload);
	    }else if ( strncasecmp(buf, "wildmat", 7) == 0 ) {
		parsecfg_wildmat(buf,f);
	    }else if ( strncasecmp(buf, "alias", 5) == 0 ) {
		parsecfg_alias(buf,f);
	    }else {
		die("Syntax error in %s on line %d", cfg.ProfileFile, linenr);
	    }
	}
    }

    fclose(f);
    
    if ( getprofile("default") == NULL )
	die("Profile \"default\" not defined");
    info("Loaded %d Profiles", master->numprofiles);

    info("Loaded %d Wildmats", master->numwildmats);

    if ( getalias("auth_default") == NULL )
	die("Alias \"auth_default\" not defined");
    if ( getalias("acct_default") == NULL )
	die("Alias \"acct_default\" not defined");
    info("Loaded %d Aliases", master->numaliases);

    stat(cfg.ProfileFile,&_stat);
    master->laststatprofile = _stat.st_mtime;
}

/*
 * Configuration parts parser functions
 */

PROTO void parsecfg_profile(char *buf, FILE *f, int reload) {
    char key[256], val[256];
    PROFILE *defp;

    key[0] = 0;
    val[0] = 0;

    if ( (sscanf(buf, "%s %s", key, val)) != 2 )
	die("Profile Syntax error on line %d of %s", linenr, cfg.ProfileFile);

    if ( reload == 0 ) {
	if ( strcmp(val, "default") != 0 ) {
	    /* we inherit (copy) from default profile */
	    if ( (defp = getprofile("default")) == NULL ) 
		die("Could not retrieve default profile, the default profile must be listed first.");

	    memcpy((profiles+master->numprofiles), defp, sizeof(PROFILE));
	}
	strcpy((profiles+master->numprofiles)->name, val);

	profile_copycfg(buf, f, (profiles+master->numprofiles));
	master->numprofiles++;
	if ( master->numprofiles > MAX_PROFILES ) 
	    die("Too many profiles (%d) increase MAX_PROFILES in nntpd.h", master->numprofiles);
    }else{
	if ( (defp = getprofile(val)) == NULL ) 
	    die("Could not find profile %s for reloading.", val);

	profile_copycfg(buf, f, defp);
    }
}

PROTO void profile_copycfg(char *buf, FILE *f, PROFILE *pf) {
    char key[256], val[256];
    int i=0;

    while( i == 0 && fgets(buf, BUFS, f) ) {
	chop(buf);

	linenr++;
	if ( buf[0] == ';' || buf[0] == '#' || buf[0] == 0 )
	    continue;

	if ( strncasecmp(buf, "end", 3) == 0 ) {
	    i++;
	    break;
	}

	key[0] = 0;
	val[0] = 0;

	if ( sscanf(buf, "%*[ \t]%64s%*[ \t]%192[^\r\n]%*[\t\r\n]", key, val) != 2 )
	    die("Error on line %d of %s (%s - %s)", linenr, cfg.ProfileFile, key, val);

	if ( strcasecmp(key, "ReadPat") == 0 ) {
	    strcpy(pf->ReadPat, val);
	}else if ( strcasecmp(key, "PostPat") == 0 ) {
	    strcpy(pf->PostPat, val);
	}else if ( strcasecmp(key, "Banner") == 0 ) {
	    strcpy(pf->Banner, val);
	}else if ( strcasecmp(key, "MaxConnections") == 0 ) {
	    pf->MaxConnections = atoi(val);
	}else if ( strcasecmp(key, "MaxConcurrent") == 0 ) {
	    pf->MaxConcurrent = atoi(val);
	}else if ( strcasecmp(key, "MaxSessionTime") == 0 ) {
	    pf->MaxSessionTime = atoi(val);
	}else if ( strcasecmp(key, "MaxSessionBytes") == 0 ) {
	    pf->MaxSessionBytes = atoi(val);
	}else if ( strcasecmp(key, "MaxBytesPerSecond") == 0 ) {
	    pf->MaxBytesPerSecond = atoi(val);
	}else if ( strcasecmp(key, "XComplaintsTo") == 0 ) {
	    strcpy(pf->XComplaintsTo, val);
	}else if ( strcasecmp(key, "Organization") == 0 ) {
	    strcpy(pf->Organization, val);
	}else if ( strcasecmp(key, "ForceOrganization") == 0 ) {
	    pf->ForceOrganization = atoi(val);
	}else if ( strcasecmp(key, "AddNNTPPostingHost") == 0 ) {
	    pf ->AddNNTPPostingHost = atoi(val);
	}else if ( strcasecmp(key, "NoControl") == 0 ) {
	    pf->NoControl = atoi(val);
	}else if ( strcasecmp(key, "PathHeader") == 0 ) {
	    strcpy(pf->PathHeader, val);
	}else if ( strcasecmp(key, "MessageIdDomain") == 0 ) {
	    strcpy(pf->MessageIdDomain, val);
	}else{
	    info("parseprofile unknown %s = %s", key, val);
	}
    }
}


PROTO void parsecfg_wildmat(char *buf, FILE *f) {
    char wmname[128], wmval[512];

    wmname[0] = 0;
    wmval[0] = 0;

    if ( sscanf(buf, "wildmat%*[ \t]%128[^ \t]%*[ \t]%512s", wmname, wmval) != 2 ) 
	die("Wildmat config syntax error on line %d of %s", linenr, cfg.ProfileFile);

    if ( (wildmats[master->numwildmats] = calloc(1,sizeof(WILDMAT)) ) == NULL )
	die("Can't allocate memory for WILDMAT");
                                                                                                                             
    wildmats[master->numwildmats]->name    = strdup(wmname);
    wildmats[master->numwildmats]->pattern = strdup(wmval);
                                                                                                                             
    master->numwildmats++;
    if ( master->numwildmats > MAX_WILDMATS ) 
	die("Too many Wildmat entries (%d) increase MAX_WILDMATS in nntpd.h", master->numwildmats);
}


PROTO void parsecfg_alias(char *buf, FILE *f) {
    char name[MAX_STRING], lib[MAX_STRING], args[MAX_STRING];

    name[0] = 0;
    lib[0] = 0;
    args[0] = 0;

    if ( sscanf(buf, "alias%*[ \t]%256[^ \t]%*[ \t]%256[^ \t]%*[ \t]%256[^\r\n]", name, lib, args) < 2 ) 
	die("Alias config syntax error on line %d of %s", linenr, cfg.ProfileFile);

    strcpy( (aliases+master->numaliases)->name, name);
    strcpy( (aliases+master->numaliases)->lib,  lib);
    strcpy( (aliases+master->numaliases)->args, args);
 
    master->numaliases++;
    if ( master->numaliases > MAX_ALIASES ) 
	die("Too many aliases (%d) increase MAX_ALIASES in nntpd.h", master->numaliases);
}

/*
 * Wildmat Functions
 */

PROTO const char* getwildmat(const char* name) {
    int i;
                                                                                                                             
    if ( name == NULL )    return default_wildmat;
    if ( name[0] == '\0' ) return default_wildmat;
                                                                                                                             
    for (i=0; i<master->numwildmats; i++)
        if ( ! strcmp(wildmats[i]->name, name) )
            return wildmats[i]->pattern;
    
    return NULL;
}

/*
 * Alias Functions
 */

PROTO ALIAS* getalias(const char *name) {
    int i;
                                                                                                                             
    for (i=0; i<master->numaliases; i++)
        if ( strcmp((aliases+i)->name, name) == 0 )
            return (aliases+i);
    
    return NULL;
}

/*
 * Profiles Functions
 */

PROTO PROFILE* getprofile(const char *name) {
    int i;

    for (i=0; i<master->numprofiles; i++)
	if ( strcmp((profiles+i)->name, name) == 0 )
	    return (profiles+i);

    return NULL;
}
